import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
from sklearn import linear_model
from sklearn.naive_bayes import GaussianNB
from sklearn import tree
from sklearn.feature_selection import chi2, f_classif, SelectKBest, mutual_info_classif
from sklearn.base import clone
from logitboost import LogitBoost
from sklearn import preprocessing
from skrebate import ReliefF
import random


#  returns auc score of model
def test_model(model, x, y, x_test, y_test):
    model.fit(x, y)
    y_hat = model.predict_proba(x_test)[:, 1]
    return roc_auc_score(y_test, y_hat)


# used for testing filler algorithms
def test_model_attributes(model, x, y, x_test, y_test, attributes_test):
    return [test_model(model,
                       x[attributes_test],
                       y,
                       x_test[attributes_test],
                       y_test), attributes_test]


with open('report.txt', 'w') as f:
    # load data, preproces it
    df = pd.read_excel("data/Hospital Treatment Factors - Region Nis Serbia 2006-2009.xlsx")
    le = preprocessing.LabelEncoder()
    df['HospitalType'] = le.fit_transform(df['HospitalType'])
    df['UrbanHousing'] = le.fit_transform(df['UrbanHousing'])
    df['Education'] = le.fit_transform(df['UrbanHousing'])
    df['Outcome'] = le.fit_transform(df['Outcome'])

    attributes = ['HospitalType', 'UrbanHousing', 'Education', 'Gender', 'Age', 'DaysofTreatment']

    # test train split
    train, test = train_test_split(df, test_size=0.2)
    train_x = train[attributes]
    train_y = train['Outcome']
    test_x = test[attributes]
    test_y = test['Outcome']

    lm = linear_model.LogisticRegression(solver="liblinear")
    lm.fit(train_x, train_y)
    print(lm.coef_, lm.intercept_)

    # getting top model
    models = [
        tree.DecisionTreeClassifier(),
        LogitBoost(),
        GaussianNB()
    ]
    top_model = sorted(models, key=lambda m: test_model(m, train_x, train_y, test_x, test_y), reverse=True)[0]

    # calculate optc and auc for top model
    optc = lm.score(test_x, test_y)
    auc_original = test_model(clone(top_model), train_x, train_y, test_x, test_y)

    if optc > 0.5 and auc_original > 0.7:
        
        relf = ReliefF(n_features_to_select=num_of_attributes)
        relf.fit(train_x.to_numpy(), train_y.to_numpy())

        # create filer algorithms
        auc_status = {'skb_chi2': {}, 'skb_f_classif': {}, 'relf': {}, 'skb_mutual_info_classif': {}}

        for num_of_attributes in range(len(attributes), 1, -1):

            # chi2
            skb_chi2 = SelectKBest(chi2, k=num_of_attributes)
            skb_chi2.fit(train_x, train_y)
            skb_chi2_attributes = skb_chi2.get_feature_names_out()
            auc_status['skb_chi2'][num_of_attributes] = test_model_attributes(clone(top_model), train_x, train_y, test_x,
                                                                              test_y, skb_chi2_attributes)

            # mutual_info_classif
            skb_mutual_info_classif = SelectKBest(mutual_info_classif, k=num_of_attributes)
            skb_mutual_info_classif.fit(train_x, train_y)
            skb_mutual_info_classif_attributes = skb_mutual_info_classif.get_feature_names_out()
            auc_status['skb_mutual_info_classif'][num_of_attributes] \
                = test_model_attributes(clone(top_model), train_x, train_y, test_x, test_y,
                                        skb_mutual_info_classif_attributes)

            # f_classif
            # skb_f_classif = SelectKBest(f_classif, k=num_of_attributes)
            # skb_f_classif.fit(train_x, train_y)
            # skb_f_classif_attributes = skb_f_classif.get_feature_names_out()
            # auc_status['skb_f_classif'][num_of_attributes] = test_model_attributes(clone(top_model), train_x, train_y,
            #                                                                        test_x, test_y, skb_f_classif_attributes)

            # relieff
            attributes_for_relf = [attributes[x] for x in relf.top_features_[:num_of_attributes]]
            auc_status['relf'][num_of_attributes] = [test_model(clone(top_model), train_x[attributes_for_relf], train_y,
                                                                test_x[attributes_for_relf], test_y), attributes_for_relf]

        # extract best filer algorithm
        max_auc = 0
        max_alg = "skb_chi2"
        max_val = 6
        max_attributes = attributes
        for alg in auc_status.keys():
            algorithm = auc_status[alg]
            for num in algorithm.keys():
                if algorithm[num][0] > max_auc:
                    max_auc = algorithm[num][0]
                    max_val = num
                    max_alg = alg
                    max_attributes = algorithm[num][1]

        f.write("Top model used for classification: " + str(top_model))
        f.write("Top filler algorithm: " + str(max_alg))
        f.write("Chosen attributes: " + str(max_attributes))
        f.write("Result for AUC algorithm: " + str(max_auc))

        f.write("Original value for AUC: " + str(auc_original))

        f.write("Prediction with K regression parameters")
        lm = linear_model.LogisticRegression(solver="liblinear")
        auc = test_model(lm, train_x[max_attributes], train_y, test_x[max_attributes], test_y)
        print(lm.coef_, lm.intercept_)
        optc = lm.score(test_x[max_attributes], test_y)
        f.write("AUC: " + str(auc))
        f.write("OPTC: " + str(optc))

        f.write("Prediction with L regression parameters")
        lm = linear_model.LogisticRegression(solver="liblinear")
        auc = test_model(lm, train_x, train_y, test_x, test_y)
        print(lm.coef_, lm.intercept_)
        optc = lm.score(test_x, test_y)
        f.write("AUC: " + str(auc))
        f.write("OPTC: " + str(optc))
    else:
        f.write("No valid prediction!")
    f.write("*****************************************")
